# Here we are creating a different dataframe where each key is representative of different locations + sequence 

import numpy as np
import pandas as pd

import matplotlib.pyplot as plt
import seaborn as sns

from pystats import stats as pyss
from msk_seg_util.util.stats_util import concordance_correlation_coefficient, calc_cv

from typing import Sequence

__all__ = ['PairedScanAnalysis']

STR_KEYS = ['Knee', 'Location', 'Subject']


class PairedScanAnalysis(object):
    """Analyze paired scan knee data.

    Analyzing paired scan data is useful for evaluating results computed with new algorithms.
    For example, one use case would be comparing the average T2 values in femoral cartilage regions, which are segmented
    manually vs automatically.

    Scans are paired per subject. Bilateral scan analysis is supported.

    Args:
        scan1_files (:obj:`Sequence[str]`): Excel files for first element in the pair to load data from. Data is loaded
            into a pandas ``DataFrame``.
        scan2_files (:obj:`Sequence[str]`): Excel files for second element in the pair to load data from. Data is loaded
            into a pandas ``DataFrame``. File order should correspond with the file order in ``scan1_files``.
        single_knee (:obj:`bool`, optional): Not bilateral. Defaults to ``False``.
        sheet_name (:obj:`str`, optional): Sheet to load from all excel files. Defaults to first sheet in each file.
    """

    def __init__(self, scan1_files, scan2_files, single_knee: bool = False, sheet_name: str = None):
        self.scan1_files = scan1_files
        self.scan2_files = scan2_files

        self.single_knee = single_knee

        self.scan1_data = self._load_data(scan1_files, single_knee=single_knee, sheet_name=sheet_name)
        self.scan2_data = self._load_data(scan2_files, single_knee=single_knee, sheet_name=sheet_name)

        self.scan1_region_data = self.reorg_data_by_region(self.scan1_data)
        self.scan2_region_data = self.reorg_data_by_region(self.scan2_data)

        scan1_region_data = self.scan1_region_data
        scan2_region_data = self.scan2_region_data

        vals_diff = scan1_region_data.drop(STR_KEYS, axis=1) - scan2_region_data.drop(STR_KEYS, axis=1)
        vals_diff[STR_KEYS] = scan1_region_data[STR_KEYS]

        vals_mean = (scan1_region_data.drop(STR_KEYS, axis=1) + scan2_region_data.drop(STR_KEYS, axis=1)) / 2
        vals_mean[STR_KEYS] = scan1_region_data[STR_KEYS]

        self.vals_diff = vals_diff
        self.vals_mean = vals_mean

    def _load_data(self, xl_files: Sequence[str], single_knee: bool = False, sheet_name: str = None):
        """Loads excel files formatted by DOSMA for femoral cartilage.

        Args:
            xl_files (Sequence[str]): Excel files to load data from. Each file corresponds to data about a different
                subject. All files should have the following column names:

                    - 'Subject': Subject id. If both knees are analyzed, this column should be of the form
                    ``<subject_id>-<R/L>``. For example,
                    - 'Side': medial/lateral compartment
                    - 'Location': deep/superficial/total
                    - 'Region': posterior/central/anterior

            single_knee (:obj:`bool`, optional): Analyze single knee.
            sheet_name (:obj:`str`, optional): Sheet to load from all excel files. Defaults to first sheet in each file.

        Returns:
            :obj:`pd.DataFrame`:
        """
        # Load excel files one by one and then cat them
        for num, fname in enumerate(xl_files):
            xls = pd.ExcelFile(fname)
            sheet_name = xls.sheet_names[0] if sheet_name is None else sheet_name
            tmp = xls.parse(sheet_name)

            if num == 0:
                data = tmp
            else:
                data = pd.concat([data, tmp])

        if data.columns.contains('Subject') is True:
            # info = data['Subject'].astype(str).str.split('-').str
            if single_knee:
                data['Knee'] = 'N/A'
            else:
                data['Subject'], data['Knee'] = data.Subject.str.split('-').str
        data.Side = data.Side.str.capitalize()
        data.Location = data.Location.str.capitalize()
        data.Region = data.Region.str.capitalize()

        return data

    def reorg_data_by_region(self, data):
        seqs = pd.DataFrame()

        # Create key per region and side
        uniq_reg = list(data.Region.unique())
        uniq_sid = list(data.Side.unique())

        for reg in uniq_reg:
            for sid in uniq_sid:
                #         Name of key and associated values
                df_name = '%s-%s' % (sid, reg)
                seqs[df_name] = data[(data.Side == sid) & (data.Region == reg)].Mean.values

        # Adding a couple of rows to keep track of subjects 
        seqs['Knee'] = data[(data.Side == sid) & (data.Region == reg)].Knee.values
        seqs['Location'] = data[(data.Side == sid) & (data.Region == reg)].Location.values
        seqs['Subject'] = data[(data.Side == sid) & (data.Region == reg)].Subject.values

        return seqs

    def _ba_plot_args(self):
        return {'ylim': (-15, 15)}

    def plot_ba_separate_knee(self, fname=None, **kwargs):
        default_args = self._ba_plot_args()
        for k in default_args.keys():
            if k in kwargs:
                default_args[k] = kwargs.get(k)

        scan1_region_data = self.scan1_region_data
        scan2_region_data = self.scan2_region_data
        vals_diff = self.vals_diff
        vals_mean = self.vals_mean

        uniq_knee = list(scan1_region_data.Knee.unique())
        uniq_loc = list(scan1_region_data.Location.unique())
        fig_cap = ['A', 'B', 'C', 'D', 'E', 'F']

        plt.close('all')
        _, axs = plt.subplots(2, 3, figsize=(35, 20))
        splot_num = 0

        column_names = list(vals_diff.columns.values)
        num_columns = 6
        cpal = sns.color_palette("bright", num_columns)

        for knee_num, knee_name in enumerate(uniq_knee):
            for cart_num, cart_reg in enumerate(uniq_loc):

                #         Isolate the values for specific cartilage region and sequence type
                diff_arr = vals_diff[(vals_diff.Location == cart_reg) & (vals_diff.Knee == knee_name)]
                diff_arr = diff_arr.drop(STR_KEYS, axis=1).values

                mean_arr = vals_mean[(vals_mean.Location == cart_reg) & (vals_mean.Knee == knee_name)]
                mean_arr = mean_arr.drop(STR_KEYS, axis=1).values

                #         Calculate the mean and std of ALL values over all regions for the BA analysis
                ba_mean = np.mean(diff_arr)
                ba_std = np.std(diff_arr)

                splot_num += 1

                for num in range(num_columns):
                    #             Plot the 6 sub-regions for each cartilage layer as well as two sequences
                    plt.subplot(2, 3, splot_num)
                    plt.plot(mean_arr[:, num], diff_arr[:, num], label=column_names[num], color=cpal[num], LineStyle="",
                             Marker="o", MarkerSize=15)

                    plt.axhline(ba_mean, color='gray', linestyle=':', Linewidth=4)
                    plt.axhline(ba_mean + 1.96 * ba_std, color='gray', LineStyle='--', LineWidth=4)
                    plt.axhline(ba_mean - 1.96 * ba_std, color='gray', LineStyle='--', LineWidth=4)

                    plt.ylabel('T2 Difference (ms)')
                    plt.ylim(default_args['ylim'])
                    plt.xlabel('T2 Mean (ms)')
                    plt.title('{}: {} T2 {} Res'.format(fig_cap[splot_num - 1], cart_reg, knee_name), fontweight='bold')
                    plt.grid(which="Major")

        plt.legend(loc=8, bbox_to_anchor=(-0.9, -0.3), fancybox=True, ncol=6)
        plt.subplots_adjust(hspace=0.3, wspace=0.3)
        if fname is not None:
            plt.savefig(fname, bbox_inches='tight')

        return axs

    def plot_ba_both_knee(self, fname=None, **kwargs):
        default_args = self._ba_plot_args()
        for k in default_args.keys():
            if k in kwargs:
                default_args[k] = kwargs.get(k)

        scan1_region_data = self.scan1_region_data
        scan2_region_data = self.scan2_region_data
        vals_diff = self.vals_diff
        vals_mean = self.vals_mean

        uniq_knee = list(scan1_region_data.Knee.unique())
        uniq_loc = list(scan1_region_data.Location.unique())
        fig_cap = ['A', 'B', 'C']

        plt.close('all')
        plt.subplots(1, 3, figsize=(35, 10))
        splot_num = 0

        column_names = list(vals_diff.columns.values)
        num_columns = 6
        cpal = sns.color_palette("bright", num_columns)

        for cart_num, cart_reg in enumerate(uniq_loc):

            #         Isolate the values for specific cartilage region and sequence type
            diff_arr = vals_diff[(vals_diff.Location == cart_reg)]
            diff_arr = diff_arr.drop(STR_KEYS, axis=1).values

            mean_arr = vals_mean[(vals_mean.Location == cart_reg)]
            mean_arr = mean_arr.drop(STR_KEYS, axis=1).values

            #         Calculate the mean and std of ALL values over all regions for the BA analysis
            ba_mean = np.mean(diff_arr)
            ba_std = np.std(diff_arr)

            splot_num += 1

            for num in range(num_columns):
                #             Plot the 6 sub-regions for each cartilage layer as well as two sequences
                plt.subplot(1, 3, splot_num)
                plt.plot(mean_arr[:, num], diff_arr[:, num], label=column_names[num], color=cpal[num], LineStyle="",
                         Marker="o", MarkerSize=15)

                plt.axhline(ba_mean, color='gray', linestyle=':', Linewidth=4)
                plt.axhline(ba_mean + 1.96 * ba_std, color='gray', LineStyle='--', LineWidth=4)
                plt.axhline(ba_mean - 1.96 * ba_std, color='gray', LineStyle='--', LineWidth=4)

                plt.ylabel('T2 Difference (ms)')
                plt.ylim(default_args['ylim'])
                plt.xlabel('T2 Mean (ms)')
                plt.title('{}: {} T2'.format(fig_cap[splot_num - 1], cart_reg), fontweight='bold')
                plt.grid(which="Major")

        plt.legend(loc=8, bbox_to_anchor=(-0.9, -0.3), fancybox=True, ncol=6)
        plt.subplots_adjust(hspace=0.3, wspace=0.3)
        if fname is not None:
            plt.savefig(fname, bbox_inches='tight')

    def print_metrics(self, skey='Location'):
        """
        Print metrics grouped by search key (skey)
        """
        s1 = self.scan1_region_data
        s2 = self.scan2_region_data

        unique_keys = list(s1[skey].unique())
        for ukey in sorted(unique_keys):
            temp1 = s1[s1[skey] == ukey].drop(STR_KEYS, axis=1).values
            temp2 = s2[s2[skey] == ukey].drop(STR_KEYS, axis=1).values

            # calculate RMS-CV
            cvs = calc_cv(temp1.flatten(), temp2.flatten())
            rms_cv = np.sqrt(np.mean((cvs) ** 2)) * 100

            # calculate concordance correlation
            cc = concordance_correlation_coefficient(temp1.flatten(), temp2.flatten())

            print('%s: %-10s \t RMS-CV: %0.2f%% \t Lin\'s Concordance: %0.3f' % (skey, ukey, rms_cv, cc))

    def kw_dunn(self, skey='Location', alpha=0.05):
        """Run Kruskal-Wallis + Dunn posthoc analysis for different regions.

        Args:
            skey (optional): Column to compare between scans. Unique values in the column will be compared across
                scan cohorts. Defaults to ``Location``.
            alpha (:obj:`float`, optional): Alpha level. Defaults to ``0.05``.
        """
        s1 = self.scan1_region_data
        s2 = self.scan2_region_data

        unique_keys = list(s1[skey].unique())
        unique_regions = s1.columns.tolist()[:6]
        print_st = ''
        print_gt = ''
        for ukey in unique_keys:
            temp1 = s1[s1[skey] == ukey].drop(STR_KEYS, axis=1).values
            temp2 = s2[s2[skey] == ukey].drop(STR_KEYS, axis=1).values
            for ureg in unique_regions:
                temp1 = s1[s1[skey] == ukey][ureg].values
                temp2 = s2[s2[skey] == ukey][ureg].values
                temp_df = pd.DataFrame([temp1, temp2])
                results = pyss.kruskal_wallis(temp_df, posthoc_test='dunn')
                results = results['dunn']
                p = np.asarray(results)[0, 1]
                res = np.absolute(np.asarray(results.values))
                if (res <= alpha).any():
                    print_st += '%s %s (p=%0.3f):\t\t1: %0.3f +/- %0.3f\t\t2: %0.3f +/- %0.3f\n' % (ukey, ureg, p,
                                                                                                    np.mean(temp1),
                                                                                                    np.std(temp1),
                                                                                                    np.mean(temp2),
                                                                                                    np.std(temp2))
                else:
                    print_gt += '%s %s (p=%0.3f)\n' % (ukey, ureg, p)
        print(print_st)
        print(print_gt)



"""Utilities for tracking metrics."""

from enum import Enum

import numpy as np
import scipy.stats as spstats
import pandas as pd

import multiprocessing as mp

from medpy.metric import dc, assd, recall, precision, sensitivity, specificity, positive_predictive_value

from typing import Iterable, Sequence, Union


def _cv(y_pred: np.ndarray, y_true: np.ndarray):
    """Calculate coefficient of variation.

    Args:
        y_true (np.ndarray): Binary ground truth mask.
        y_pred (np.ndarray): Predicted ground truth mask.

    Returns:
        float: Coefficient of variation.
    """
    y_true = np.squeeze(y_true)
    y_pred = np.squeeze(y_pred)

    cv = np.std([np.sum(y_true), np.sum(y_pred)]) / np.mean([np.sum(y_true), np.sum(y_pred)])
    return cv


def _volumetric_overlap_error(y_pred: np.ndarray, y_true: np.ndarray):
    """Volumetric overlap error.

    Args:
        y_true (np.ndarray): Binary ground truth mask.
        y_pred (np.ndarray): Predicted ground truth mask.

    Returns:
        float: Volumetric overlap error.
    """

    y_true = y_true.flatten()
    y_pred = y_pred.flatten()

    y_true_bool = np.asarray(y_true, dtype=np.bool)
    y_pred_bool = np.asarray(y_pred, dtype=np.bool)
    TP = np.sum(y_true_bool * y_pred_bool, axis=-1)
    FP = np.sum(~y_true_bool * y_pred_bool, axis=-1)
    FN = np.sum(y_true_bool * ~y_pred_bool, axis=-1)

    mu = 1e-07

    voe = 1 - (TP + mu) / (TP + FP + FN + mu)

    return voe


class MetricOperation(Enum):
    """Operation to calculate group statistics.

    To evaluate the overall performance on a group, results for each member of the group is averaged.
    However, in many cases, this may not be the best option.

    The following operations are supported:

        1. MEAN
        2. MEDIAN
        3. RMS: Root mean square.
    """
    MEAN = 1, lambda x, **kwargs: np.mean(x, **kwargs)
    MEDIAN = 2, lambda x, **kwargs: np.median(x, **kwargs)
    RMS = 3, lambda x, **kwargs: np.sqrt(np.mean(x ** 2, **kwargs))

    def __new__(cls, keycode, func):
        obj = object.__new__(cls)
        obj._value_ = keycode
        obj.func = func
        return obj

    def compute(self, x: np.ndarray, **kwargs):
        """Compute operation.

        Args:
            x (np.ndarray): Array-like over which to compute operation.
            **kwargs: Keywords for numpy implementation of mean/median.
        """
        return self.func(x, **kwargs)


def __sem__(x, **kwargs):
    """Calculate standard error of the mean."""
    args = {'axis': 0, 'ddof': 0}
    args.update(**kwargs)

    return spstats.sem(x, **args)


class MetricError(Enum):
    """Error metrics.

        1. STANDARD_DEVIATION
        2. STANDARD_ERROR: Standard error of the mean.
    """
    STANDARD_DEVIATION = 1, lambda x, **kwargs: np.std(x, **kwargs)
    STANDARD_ERROR = 2, lambda x, **kwargs: __sem__(x, **kwargs)

    def __new__(cls, keycode, func):
        obj = object.__new__(cls)
        obj._value_ = keycode
        obj.func = func
        return obj

    def compute(self, x: np.ndarray, **kwargs):
        return self.func(x, **kwargs)


class SegMetric(Enum):
    """Standard metrics used to judge segmentation quality.

    Following metrics are supported:

        1. `DSC`: Dice score coefficient.
        2. `VOE`: Volumetric overlap error.
        3. `CV`: Coefficient of variation.
        4. `ASSD`: Average Symmetric Surface Distance
        5. `Precision`
        6. `Recall`
        7. `SENSITIVITY`
        8. `SPECIFICITY`
        9. `PPV`: Positive Predictive Value

    Note `ASSD` takes long time to compute.
    """
    DSC = 1, 'Dice Score Coefficient', dc, False
    VOE = 2, 'Volumetric Overlap Error', _volumetric_overlap_error, False
    CV = 3, 'Coefficient of Variation', _cv, False
    ASSD = 4, 'Average Symmetric Surface Distance', assd, True
    PRECISION = 5, 'Precision', precision, False
    RECALL = 6, 'Recall', recall, False
    SENSITIVITY = 7, 'Sensitivity', sensitivity, False
    SPECIFICITY = 8, 'Specificity', specificity, False
    PPV = 9, "Positive Predictive Value", positive_predictive_value, False

    def __new__(cls, key_code, full_name, func, use_voxel_spacing=False):
        obj = object.__new__(cls)
        obj._value_ = key_code
        obj.full_name = full_name
        obj.func = func
        obj.use_voxel_spacing = use_voxel_spacing
        return obj

    def compute(self, y_true, y_pred, voxel_spacing):
        """Compute metric for a single segmentation.

        Args:
            y_true (np.ndarray): Binary ground truth mask.
            y_pred (np.ndarray): Predicted ground truth mask.
            voxel_spacing (tuple[float]): Voxel dimensions (in mm).

        Returns:
            The computed metric.
        """
        if self.use_voxel_spacing:
            return self.func(y_pred, y_true, voxel_spacing)
        else:
            return self.func(y_pred, y_true)


class SegMetricsProcessor(object):
    """This class processes, handles, and stores common segmentation metrics.

    Args:
        metrics (Sequence[SegMetric]): Metrics to analyze.
        class_names (Sequence[str]): Names of different classes to segment, in order of input channels.
        metric_computations (Sequence[MetricOperation], optional): Operations to perform on all metrics.
            Defaults to :obj:`(MetricOperation.MEAN, MetricOperation.RMS, MetricOperation.MEDIAN,
            MetricError.STANDARD_DEVIATION)`.
        error_metric (MetricError): Defaults to :obj:`MetricError.STANDARD_DEVIATION`.
    """

    # Default is to capitalize all metric names. If another name is, please specify here.
    __METRICS_DISPLAY_NAMES = {SegMetric.DSC: SegMetric.DSC.name,
                               SegMetric.VOE: SegMetric.VOE.name,
                               SegMetric.CV: SegMetric.CV.name,
                               SegMetric.ASSD: 'ASSD (mm)',
                               SegMetric.PRECISION: 'Precision',
                               SegMetric.RECALL: 'Recall',
                               SegMetric.SENSITIVITY: 'Sensitivity',
                               SegMetric.SPECIFICITY: 'Specificity',
                               SegMetric.PPV: SegMetric.PPV.name}

    # Default is to calculate mean for every metric expect coefficient of variation (CV),
    # which uses root-mean-squared (RMS).
    __DEFAULT_METRIC_COMPUTATIONS = (MetricOperation.MEAN,
                                     MetricOperation.RMS,
                                     MetricOperation.MEDIAN,
                                     MetricError.STANDARD_DEVIATION,
                                     )

    def __init__(self, metrics: Sequence[SegMetric], class_names: Sequence[str],
                 metric_computations: Sequence[MetricOperation] = __DEFAULT_METRIC_COMPUTATIONS,
                 error_metric: MetricError = MetricError.STANDARD_DEVIATION):
        self.metrics = metrics
        self.__metric_names = [self.__METRICS_DISPLAY_NAMES[m] for m in self.metrics]

        self.class_names = class_names

        self.__scan_ids = []
        self.__scan_seg_data = dict()
        self.__data = dict()
        self.__is_data_stale = False

        # Default computations to perform on each metric.
        self.metric_computations = metric_computations
        self.error_metric = error_metric

        self.__dataframe = None

    def _compute_metrics(self, y_true: np.ndarray, y_pred: np.ndarray, voxel_spacing: tuple):
        """Internal helper for computing metrics.

        Args:
            y_true (np.ndarray): Binary ground truth mask.
            y_pred (np.ndarray): Predicted ground truth mask.
            voxel_spacing (tuple[float]): Voxel dimensions (in mm).

        Returns:
            pd.DataFrame: A DataFrame indexed by class names (rows) and spanned by metrics (columns).

            TODO: Draw table
        """
        assert type(y_true) is np.ndarray and type(y_pred) is np.ndarray, "y_true and y_pred must be numpy arrays"
        assert y_true.shape == y_pred.shape, "Shape mismatch: y_true and y_pred must have the same shape"
        assert y_true.ndim == 3 or y_true.ndim == 4, "Arrays must be (Y,X,Z) or (Y, X, Z, #classes)"

        if y_true.ndim == 3:
            y_true = y_true[..., np.newaxis]
            y_pred = y_pred[..., np.newaxis]

        assert y_true.shape[-1] == len(self.class_names), "Expected %d classes. Got %d" % (len(self.class_names),
                                                                                           y_true.shape[-1])
        num_classes = len(self.class_names)

        metrics_data = []
        for m in self.metrics:
            metrics_data.append([m.compute(y_true[..., c], y_pred[..., c], voxel_spacing) for c in range(num_classes)])

        return pd.DataFrame(metrics_data, index=self.__metric_names, columns=self.class_names)

    def _to_dataframe(self, scan_ids):
        """Internal helper to convert metric data into dataframe.

        Format:

              ID |                 Scan1 | Scan2 | Scan3 | Scan4 | Scan5 | ...
            ------------------------------------------------------------------
            Metric     Class
            ------------------------------------------------------------------
            Metric1    Class1
            Metric1    Class2
               .        .
               .        .
               .        .
            Metric2    Class1
            Metric2    Class2
               .        .
               .        .
               .        .

        Args:
            scan_ids: Scan ids to organize in dataframe.

        Returns:
            pd.DataFrame: A pandas Dataframe.
        """
        scan_data = []

        for scan_id in scan_ids:
            df = self.__scan_seg_data[scan_id]
            df.insert(0, 'ID', [scan_id] * len(df))
            scan_data.append(df)

        df = pd.concat(scan_data)
        df.insert(1, 'Metric', df.index)
        df = pd.melt(df, id_vars=['ID', 'Metric'], var_name='Class')
        df = df.reset_index(drop=True)
        df = df.pivot_table(index=['Metric', 'Class'], columns='ID')
        df.columns = df.columns.droplevel(0)

        return df

    def _unroll_compute_metrics(self, data_args):
        """Wrapper for computing metrics.

        Args:
            data_args (tuple): Tuple of (:obj:`scan_id`, :obj:`y_true`, :obj:`y_pred`, :obj:`voxel_spacing`).
        """
        return self._compute_metrics(data_args[1], data_args[2], data_args[3])

    def batch_compute_metrics(self, data: Iterable, num_workers: int=0):
        """Compute segmentation metrics for volume in batch using multiprocessing library.

        Args:
            data (:obj:`Iterable[tuple]`): An iterable of tuples of ``(scan_id, y_true, y_pred, voxel_spacing)``
            num_workers (:obj:`int`, optional): Number of workers to use. Defaults to `0`, meaning processes will be run
                on main thread with single worker.
        """
        scan_ids = [d[0] for d in data]
        overlapping_scan_ids = set(self.__scan_ids).intersection(set(scan_ids))
        if len(overlapping_scan_ids) != 0:
            raise ValueError('Scan id(s) %s exist.' % overlapping_scan_ids)

        with mp.Pool(num_workers) as p:
            results = p.map(self._unroll_compute_metrics, data)

        for ind, result in enumerate(results):
            scan_id = scan_ids[ind]
            self.__scan_ids.append(scan_id)
            self.__scan_seg_data[scan_id] = result

    def compute_metrics(self, scan_id: Union[str, int], y_true: np.ndarray, y_pred: np.ndarray, voxel_spacing: tuple):
        """Compute segmentation metrics for volume.

        Args:
            scan_id (:obj:`Union[str, int]`): The scan id.
            y_true (:obj:`np.ndarray`): 3D or 4D ground truth binary mask of shape ``(Y, X, Z, [# Classes])``.
            y_pred (np.ndarray): 3D or 4D prediction binary mask of shape ``(Y, X, Z, [# Classes])``.
            voxel_spacing (tuple[float]): Voxel dimensions (in mm).

        Returns:
            str: Summary of scan metrics.

        Raises:
            ValueError: If ``scan_id`` exists.
        """
        if scan_id in self.__scan_ids:
            raise ValueError('Scan id already exists, use different id')

        metrics_data = self._compute_metrics(y_true, y_pred, voxel_spacing)

        self.__scan_ids.append(scan_id)
        self.__scan_seg_data[scan_id] = metrics_data
        self.__is_data_stale = True

        return self.scan_summary(scan_id)

    def scan_summary(self, scan_id: Union[str, int], sep: str= ','):
        """Provide summary of segmentation metrics for a specific sample.

        Metrics are summarized by averaging metrics across all classes. To get the raw matrix, use
        :obj:`self.scan_data(id)`.

        Args:
            scan_id (Union[str, int]): The scan id.
            sep (str): Delimiter between entries. Defaults to ``''``.

        Returns:
            str: Performance summary for scan.
        """
        scan_data = self.__scan_seg_data[scan_id]
        avg_data = scan_data.mean(axis=1)

        metrics = avg_data.index.tolist()

        basic_format = '%s: %0.3f' + ('%s ' % sep)
        summary_str_format = basic_format * len(metrics)
        summary_str_format = summary_str_format[:-2]

        data = []
        for name in avg_data.index.tolist():
            data.extend([name, avg_data[name]])

        return summary_str_format % (tuple(data))

    def summary(self):
        computation_result = []
        df = self.dataframe
        for m_computation in self.metric_computations:
            cr = m_computation.compute(df, axis=1)
            if not isinstance(cr, pd.DataFrame):
                cr = pd.DataFrame(cr, index=df.index)
            cr.columns = [m_computation.name]
            computation_result.append(cr)

        return pd.concat(computation_result, axis=1)

    def scan_data(self, scan_id: Union[str, int]) -> pd.DataFrame:
        """Get performance table for a scan.

        Args:
             scan_id (Union[str, int]): The scan id.

        Returns:
            pd.DataFrame: Performance table for scan.
        """
        return self.__scan_seg_data[scan_id]

    @property
    def dataframe(self):
        """pd.DataFrame: Metrics as DataFrame.

        The dataframe is indexed first by the metric, then by the class. The columns correspond to the different
        samples.

        Format:

              ID                  Scan1   Scan2   Scan3   Scan4   Scan5   ...
            Metric     Class
            ------------------------------------------------------------------
            Metric1    Class1
            Metric1    Class2
               .        .
               .        .
               .        .
            Metric2    Class1
            Metric2    Class2
               .        .
               .        .
               .        .
        """
        if self.__dataframe is None or self.__is_data_stale:
            self.__dataframe = self._to_dataframe(self.__scan_ids)
            self.__is_data_stale = False

        return self.__dataframe



import os

import pickle
import h5py


def create_knee_dict(base_path):
    subject_base_str = 'rep_%02d'  # e.g.'rep_03', from 2-7
    knees = ['LEFT', 'RIGHT']

    knee_dict = {}
    # initialize map from subject-knee --> series
    for sub in range(2, 8):
        for knee in knees:
            curr_sub = subject_base_str % sub
            k = '%s-%s' % (curr_sub, knee)
            series = []
            sub_knee_dir = os.path.join(base_path, curr_sub, knee)
            files = os.listdir(sub_knee_dir)
            for f in files:
                if os.path.isdir(os.path.join(sub_knee_dir, f)):
                    series.append(f)
            series.sort()

            knee_dict[k] = (curr_sub, knee, series)

    return knee_dict


def mkdirs(dir_path: str):
    """Recursively create directories if directory does not exist.

    Args:
        dir_path (str): Directory to create. If exists, no intermediate directories will be created.

    Returns:
        str: ``dir_path``.

    Examples:
        >>> mkdirs('/example/directory')
        '/example/directory'
    """
    if not os.path.isdir(dir_path):
        os.makedirs(dir_path)

    return dir_path


def load_h5(file_path: str):
    """Load data in H5DF format.

    Args:
        file_path (str): Path to H5 file.

    Returns:
        dict: Data dictionary.

    Raises:
        FileNotFoundError: If ``file_path`` doesn't exist.
    """
    if not os.path.isfile(file_path):
        raise FileNotFoundError("{} does not exist".format(file_path))

    data = dict()
    with h5py.File(file_path, "r") as f:
        for key in f.keys():
            data[key] = f.get(key).value

    return data


def save_pik(data, file_path: str):
    """Save data using ``pickle``.

    Args:
        data: Data to save.
        file_path (str): File path.
    """
    with open(file_path, "wb") as f:
        pickle.dump(data, f)


def load_pik(file_path: str):
    """Load data using ``pickle``.

    Args:
        file_path (str): File path.

    Returns:
        Data saved using :obj:`save_pik`
    """
    with open(file_path, "rb") as f:
        return pickle.load(f)


def save_ims(file_path):
    im_path = '%s.im' % file_path
    with h5py.File(im_path, 'r') as f:
        im = f['data'][:]

    seg_path = '%s.seg' % file_path
    with h5py.File(seg_path, 'r') as f:
        seg = f['data'][:].astype('float32')
        seg = seg[..., 0, 0]
    file_path = mkdirs(file_path)
    # save ims
    cv2.imwrite(os.path.join(file_path, 'im.png'), scale_img(im))

    # save segs
    cv2.imwrite(os.path.join(file_path, 'seg.png'), scale_img(seg))


import numpy as np


def concordance_correlation_coefficient(y_true, y_pred,
                                        sample_weight=None,
                                        multioutput='uniform_average'):
    """Concordance correlation coefficient.

    The concordance correlation coefficient is a measure of inter-rater agreement.
    It measures the deviation of the relationship between predicted and true values
    from the 45 degree angle.

    Read more: https://en.wikipedia.org/wiki/Concordance_correlation_coefficient

    Original paper: Lawrence, I., and Kuei Lin. "A concordance correlation coefficient to evaluate reproducibility."
    Biometrics (1989): 255-268.

    Parameters:
        y_true : array-like of shape = (n_samples) or (n_samples, n_outputs)
            Ground truth (correct) target values.
        y_pred : array-like of shape = (n_samples) or (n_samples, n_outputs)
            Estimated target values.

    Returns:
        float: Loss in the range [-1,1].

        A value of 1 indicates perfect agreement between the true and the predicted values.

    Examples:

        >>> y_true = [3, -0.5, 2, 7]
        >>> y_pred = [2.5, 0.0, 2, 8]
        >>> concordance_correlation_coefficient(y_true, y_pred)
        0.97678916827853024

    """
    cor = np.corrcoef(y_true, y_pred)[0][1]

    mean_true = np.mean(y_true)
    mean_pred = np.mean(y_pred)

    var_true = np.var(y_true)
    var_pred = np.var(y_pred)

    sd_true = np.std(y_true)
    sd_pred = np.std(y_pred)

    numerator = 2 * cor * sd_true * sd_pred

    denominator = var_true + var_pred + (mean_true - mean_pred) ** 2

    return numerator / denominator


def calc_cv(y_true, y_pred):
    assert y_true.shape == y_pred.shape, "shape mismatch"
    assert y_true.ndim == 1, "all arrays must be flattened"

    cvs = np.std([y_true, y_pred], axis=0) / np.mean([y_true, y_pred], axis=0)

    return cvs



import sys
from setuptools import setup, find_packages

if sys.version_info < (3, 6):
    sys.exit('Sorry, Python < 3.6 is not supported')

REQUIRED_PACKAGES = ['h5py',
                     'matplotlib',
                     'medpy',
                     'numpy',
                     'opencv-python',
                     'pandas',
                     'pystats',
                     'scipy',
                     'scikit-posthocs',
                     'seaborn',
                     ]

with open("README.rst", "r") as f:
    long_description = f.read()

setup(name='msk_seg_util',
      version='0.0.1',
      description='Python package for MSK segmentation utilities',
      long_description=long_description,
      long_description_content_type="text/x-rst",
      url='https://github.com/ad12/msk_seg_util',
      author='Arjun Desai',
      author_email='arjundd@stanford.edu',
      packages=find_packages(),
      install_requires=REQUIRED_PACKAGES,
      classifiers=[
          "Programming Language :: Python :: 3",
          "Operating System :: OS Independent",
      ],
      license='MIT',
      python_requires='>=3.6',
)



import os

import cv2
import numpy as np
import seaborn as sns

from msk_seg_util.util.io_util import mkdirs


class MultiClassOverlay(object):
    """This class overlays masks on volumes.

    Args:
        num_classes (int): Number of classes, including background.
        color_palette: Sequence of ``(r,g,b)`` colors. Defaults to ``sns.color_palette('bright')``.
            For example, ``[(0.5, 0, 0), (0.11, 0.45, 0.67)]``.
        background_label (int): Voxels with this value will be designated as background. Either ``0`` or ``None``.
            To include color for background, set to ``None``. Defaults to ``0``.
        opacity (float): Color overlay opacity. Must be in interval ``[0, 1]``. Defaults to ``0.7``.

    Attributes:
        num_classes (int): Number of classes, including background.
        background_label (int): Voxels with this value will be designated as background. Either ``0`` or ``None``.
        opacity (float): Color overlay opacity.
    """

    def __init__(self, num_classes,
                 color_palette=None,
                 background_label=0,
                 opacity=0.7):
        effective_num_classes = num_classes - 1 if background_label is not None else num_classes
        color_palette = sns.color_palette("bright") if color_palette is None else color_palette

        if len(color_palette) < effective_num_classes:
            raise ValueError("Must provide at least {} colors".format(effective_num_classes))

        if opacity < 0 or opacity > 1:
            raise ValueError("`opacity` must be in range [0, 1]")

        self.num_classes = num_classes
        self.background_label = background_label
        self.opacity = opacity

        # set colormap
        color_palette = color_palette
        colormap = dict()
        cp_ind = 0
        for i in range(num_classes):
            if i == background_label:
                continue
            cp_ind += 1
            colormap[i] = color_palette[cp_ind]
        self.colormap = colormap

    def overlay(self, volume: np.ndarray, masks: np.ndarray, save_path: str = None):
        """Overlay volume with labels.

        Voxels labeled as multiple classes will be colored with the last corresponding class in the ``masks`` array.
        To avoid this, make sure voxels are one-hot-encoded and have no overlap between masks.

        Args:
            volume (:obj:`np.ndarray`): 3D volume that will be colored. Must be grayscale.
            masks (:obj:`np.ndarray`): 4D binary mask with shape ``[Y, X, Z, # classes]``.
            save_path (:obj:`str`, optional): If specified, 2D images will be saved to directory.

        Returns:
            :obj:`np.ndarray`: 3D colored volume.
        """
        if volume.ndim != 3:
            raise ValueError("Volume must be 3D array with shape [Y, X, Z].")
        if masks.ndim != 4:
            raise ValueError("Masks must be 4D binary array with shape [Y, X, Z, classes].")

        # Labels are argmax(masks) in the class dimension.
        labels = self._masks_to_labels(masks)
        labels_colored = self._apply_colormap(labels)

        vol_rgb = np.zeros(volume.shape + (3,))
        for z in range(volume.shape[-1]):
            x_im = volume[..., z]
            label_overlay = labels_colored[..., z, :]

            slice_name = '%03d.png' % (z + 1)

            filepath = os.path.join(mkdirs(save_path), slice_name) if save_path is not None else None
            im_rgb = self._im_overlay(x_im, label_overlay, filepath)
            vol_rgb[..., z, :] = im_rgb

        return vol_rgb

    def _masks_to_labels(self, masks: np.ndarray):
        """Convert 4D masks to 3D label array.

        Args:
            masks (:obj:`np.ndarray`): 4D binary mask with shape ``(Y, X, Z, N)``.
                ``N`` corresponds to number of classes.

        Returns:
            np.ndarray: 3D array of class ids.

            Voxels values corresponding to class id. Non-background classes are 1-indexed. ``0`` corresponds to
            background voxel.
        """
        assert masks.ndim == 4, "Masks must be 4D binary array with shape (Y, X, Z, # classes)"
        masks = np.array(masks)
        for i in range(0, masks.shape[-1]):
            masks[..., i] *= (i + 1)

        return np.max(masks, axis=-1)

    def _apply_colormap(self, labels: np.ndarray):
        """Converts labels into color array.

        Args:
            labels (:obj:`np.ndarray`): 3D array of class ids.
        """
        colormap = self.colormap
        background_label = self.background_label

        labels_colored = np.zeros(labels.shape + (3,))

        for c in np.unique(labels):
            if c == background_label:
                continue

            labels_colored[labels == c, :] = colormap[c]

        return (labels_colored * 255).astype(np.uint8)

    def _im_overlay(self, x: np.ndarray, c_label: np.ndarray, out_file: str=None):
        """Color 2D grayscale image with labels.

        Args:
            x (:obj:`np.ndarray`): 2D grayscale image with shape ``(Y, X)``.
            c_label (:obj:`np.ndarray`): 2D rgb image with shape ``(Y, X, C)``.
            out_file (:obj:`str`, optional): If specified, image written to file.

        Returns:
            np.ndarray: 2D rgb image. Type: uint8.
        """
        x_o = scale_img(np.squeeze(x))
        x_rgb = np.stack([x_o, x_o, x_o], axis=-1).astype(np.uint8)
        overlap_img = cv2.addWeighted(x_rgb, 1, c_label, self.opacity, 0)

        if out_file:
            cv2.imwrite(out_file, cv2.cvtColor(overlap_img, cv2.COLOR_RGB2BGR))

        return overlap_img.astype(np.uint8)


def write_sep_im_overlay(dir_path, xs, y_true, y_pred):
    """Overlap input (xs) with mask (im_overlap) and save to directory.

    :param dir_path: path to directory to save images
    :param xs: inputs
    :param im_overlay: overlay images
    """
    correct_dir_path = mkdirs(os.path.join(dir_path, 'true_pos'))
    error_dir_path = mkdirs(os.path.join(dir_path, 'error'))
    num_slices = xs.shape[0]
    for i in range(num_slices):
        x = scale_img(np.squeeze(xs[i, ...]))
        x = np.stack([x, x, x], axis=-1).astype(np.uint8)
        im_correct, im_error = generate_sep_ovlp_image(y_true[i, ...], y_pred[i, ...])

        slice_name = '%03d.png' % i

        overlap_img_correct = cv2.addWeighted(x, 1, im_correct, 1.0, 0)
        cv2.imwrite(os.path.join(correct_dir_path, slice_name), overlap_img_correct)

        overlap_img_error = cv2.addWeighted(x, 1, im_error, 1.0, 0)
        cv2.imwrite(os.path.join(error_dir_path, slice_name), overlap_img_error)


def generate_sep_ovlp_image(y_true, y_pred):
    """
    TODO: write comment
    :param y_true: numpy array of ground truth labels
    :param y_pred: numpy array of predicted labels
    :return: a BGR image
    """
    assert (y_true.shape == y_pred.shape)
    assert len(y_true.shape) == 2, "shape should be 2d, but is " + y_true.shape

    y_true = y_true.astype(np.bool)
    y_pred = y_pred.astype(np.bool)

    TP = y_true * y_pred
    FN = y_true * (~y_pred)
    FP = (~y_true) * y_pred

    # BGR format
    img_corr = np.stack([np.zeros(TP.shape), TP, np.zeros(TP.shape)], axis=-1).astype(np.uint8) * 255
    img_err = np.stack([FP, np.zeros(TP.shape), FN], axis=-1).astype(np.uint8) * 255

    return (img_corr, img_err)


def write_ovlp_masks(dir_path, y_true, y_pred):
    """
    Overlap ground truth with prediction and save to directory
    Red - false negative
    Green - true positive
    Blue - false negative
    :param dir_path: path to directory
    :param y_true: numpy array of ground truth labels
    :param y_pred: numpy array of predicted labels
    """
    dir_path = mkdirs(dir_path)
    y_true = np.squeeze(y_true)
    y_pred = np.squeeze(y_pred)

    assert (y_true.shape == y_pred.shape)
    ims = []
    num_slices = y_true.shape[0]
    for i in range(num_slices):
        slice_true = y_true[i, :, :]
        slice_pred = y_pred[i, :, :]
        img = make_accuracy_map(slice_true, slice_pred)
        ims.append(img)

        slice_name = '%03d.png' % i
        cv2.imwrite(os.path.join(dir_path, slice_name), img)

    return ims


def write_mask(dir_path, y_true):
    """
    Save ground truth mask to directory
    :param dir_path: path to directory
    :param y_true: numpy array of ground truth labels
    """
    dir_path = mkdirs(dir_path)
    y_true = np.squeeze(y_true) * 255
    num_slices = y_true.shape[0]
    for i in range(num_slices):
        slice_name = '%03d.png' % i
        cv2.imwrite(os.path.join(dir_path, slice_name), y_true[i, :, :])


def write_prob_map(dir_path, y_probs):
    """
    Write probablity map for prediction as image (colormap jet)
    :param dir_path: path to directory
    :param y_probs: numpy array of prediction probabilities
    """
    dir_path = mkdirs(dir_path)
    y_probs = np.squeeze(y_probs)
    num_slices = y_probs.shape[0]
    for i in range(num_slices):
        slice_name = '%03d.png' % i
        im = y_probs[i, :, :] * 255
        im = im[..., np.newaxis].astype(np.uint8)
        imC = cv2.applyColorMap(im, cv2.COLORMAP_HOT)
        cv2.imwrite(os.path.join(dir_path, slice_name), imC)


def scale_img(im, scale=255):
    """
    Scale image from 0-scale
    :param im: input image
    :param scale: max value
    :return:
    """
    im = im.astype(np.float32)
    im = im - np.min(im)
    im = im / np.max(im)
    im *= scale

    return im


def make_accuracy_map(y_true: np.ndarray, y_pred: np.ndarray, only_err: bool = False):
    """Make rgb map of true positives, false positives, and false negatives.

    True positives are green, false positives are red, false negatives are blue.

    Args:
        y_true (:obj:`np.ndarray`): Ground truth binary mask(s).
        y_pred (:obj:`np.ndarray`): Prediction binary mask(s).
        only_err (:obj:`bool`, optional): Only map errors - false positives and false negatives.
            Defaults to ``False``.

    Returns:
        :obj:`np.ndarray`: A rgb array. Will have extra channel dimension.
    """
    if y_true.shape != y_pred.shape:
        raise ValueError("`y_true` and `y_pred` must have same shapes")
    assert len(y_true.shape) == 2, "shape should be 2d, but is {}".format(y_true.shape)

    y_true = y_true.astype(np.bool)
    y_pred = y_pred.astype(np.bool)

    TP = np.zeros(y_true.shape) if only_err else y_true & y_pred
    FN = y_true & (~y_pred)
    FP = (~y_true) & y_pred

    # RGB format.
    img = np.stack([FP, TP, FN], axis=-1).astype(np.uint8)

    return img



import numpy as np
import matplotlib.pyplot as plt


def plot_bland_altman(tp1, tp2, metric='', ylim=None):
    data_diff = tp1 - tp2
    data_mean = (tp1 + tp2) / 2

    ba_mean = np.mean(data_diff)
    ba_std  = np.std(data_diff)

    plt.close('all')
    plt.subplots(1,1, figsize=(10,10))
    plt.subplot(1,1,1)
    plt.plot(data_mean, data_diff, LineStyle = "", Marker="o", MarkerSize = 15)

    plt.axhline(ba_mean, color='gray', linestyle=':', Linewidth = 4)
    plt.axhline(ba_mean + 1.96*ba_std, color='gray', LineStyle='--', LineWidth = 4)
    plt.axhline(ba_mean - 1.96*ba_std, color='gray', LineStyle='--', LineWidth = 4)    

    plt.ylabel('%s Difference (%%)' % metric)                
    plt.xlabel('%s Mean (%%)' % metric)
    plt.title('%s Scan/Re-scan' % metric, fontweight='bold')
    plt.grid(which="Major")
    
    if ylim:
        plt.ylim(ylim)


def portfolio(ims, ims_overlays=None, cmap='hot', alpha=0.5, figsize=(30,30)):
    """
    Output portfolio of images as subplots
    param ims: 1D/2D list of RGB numpy arrays to display
    """
    is_1d = type(ims[0]) is not list
    nrows = len(ims) if not is_1d else 1
    ncols = len(ims[0]) if not is_1d else len(ims)
    
    fig, axs = plt.subplots(nrows=nrows, ncols=ncols, dpi=100, figsize=figsize)

    for r in range(nrows):
        for c in range(ncols):
            ax = axs[r, c] if not is_1d else axs[c]
            im = ims[r, c] if not is_1d else ims[c]
            ax.imshow(im)
            if ims_overlays is not None:
                imo = ims_overlays[r, c] if not is_1d else ims_overlays[c]
                ax.imshow(im, cmap=cmap, alpha=alpha)
                          
            ax.axis('equal')
    
    fig.tight_layout()
    plt.show()


def scale(x, ub=1):
    xo = x - np.min(x)
    xo = xo / np.max(xo)
    xo *= ub
    
    return xo



import unittest

import numpy as np

from msk_seg_util.util.metric_util import SegMetric
from msk_seg_util.util.metric_util import SegMetricsProcessor


class TestSegMetricsProcessor(unittest.TestCase):
    _MASK = np.asarray(
        [
            [
                [[1, 1], [0, 0]],
                [[0, 1], [0, 0]],
            ],
            [
                [[1, 1], [1, 1]],
                [[1, 1], [1, 1]],
            ],
            [
                [[1, 0], [0, 1]],
                [[1, 0], [0, 0]],
            ],
            [
                [[0, 0], [0, 1]],
                [[0, 0], [0, 0]],
            ]

        ])
    _MASK = np.transpose(_MASK, (1, 2, 3, 0))

    _PREDICTION = np.asarray(
        [
            [
                [[0, 1], [1, 0]],
                [[0, 1], [1, 0]],
            ],
            [
                [[0, 0], [0, 1]],
                [[1, 1], [0, 1]],
            ],
            [
                [[1, 1], [1, 1]],
                [[1, 1], [1, 0]],
            ],
            [
                [[0, 1], [1, 1]],
                [[0, 0], [0, 0]],
            ]

        ])
    _PREDICTION = np.transpose(_PREDICTION, (1, 2, 3, 0))

    _CLASS_NAMES = ('Class-1', 'Class-2', 'Class-3', 'Class-4')
    _VOXEL_SPACING = (1, 1, 1)
    _METRICS = [SegMetric.DSC, SegMetric.VOE, SegMetric.CV, SegMetric.ASSD, SegMetric.PRECISION, SegMetric.RECALL,
                SegMetric.SENSITIVITY, SegMetric.SPECIFICITY, SegMetric.PPV]

    def test_batch_compute_metrics(self):
        manager = SegMetricsProcessor(metrics=self._METRICS, class_names=self._CLASS_NAMES)

        scan_data = []
        for i in range(10):
            scan_data.append(('Scan%d' % i, self._MASK, self._PREDICTION, self._VOXEL_SPACING))

        manager.batch_compute_metrics(scan_data, num_workers=2)

    def test_compute_metrics(self):
        manager = SegMetricsProcessor(metrics=self._METRICS, class_names=self._CLASS_NAMES)
        manager.compute_metrics('Scan1', self._MASK, self._PREDICTION, self._VOXEL_SPACING)
        manager.compute_metrics('Scan2', self._MASK, self._PREDICTION, self._VOXEL_SPACING)


if __name__ == '__main__':
    unittest.main()
